import pytest
import numpy as np
import lightgbm as lgb
from arfs.feature_selection.allrelevant import Leshy, BoostAGroota, GrootCV
from arfs.utils import (
    generated_corr_dataset_regr,
    generated_corr_dataset_classification,
)
from arfs.utils import LightForestClassifier, LightForestRegressor


class TestLeshy:
    """
    Test suite for all-relevant FS boruta-like method: Leshy
    """

    def test_borutaPy_vs_leshy_with_rfc_and_native_feature_importance(self):
        # too slow for circleci to run them in a reasonable time
        # takes 2 min on laptop, 1h or more on circleci
        # sklearn random forest implementation
        # X, y, w = _generated_corr_dataset_classification()
        # rfc = RandomForestClassifier(max_features='sqrt', max_samples=0.632, n_estimators=100)
        # bt = BorutaPy(rfc)
        # bt.fit(X.values, y)
        # borutapy_rfc_list = sorted(list(X.columns[bt.support_]))

        # lightGBM random forest implementation
        baseline_list = ["var0", "var1", "var2", "var3", "var4"]
        X, y, w = generated_corr_dataset_classification(size=100)
        n_feat = X.shape[1]
        rfc = LightForestClassifier(n_feat)
        # RandomForestClassifier(max_features='sqrt', max_samples=0.632, n_estimators=100) # --> too slow
        arfs = Leshy(rfc, verbose=0, max_iter=10, random_state=42, importance="native")
        arfs.fit(X, y)
        leshy_rfc_list = sorted(arfs.support_names_)

        # assert borutapy_rfc_list == leshy_rfc_list, "same selected features are expected"
        assert bool(
            set(baseline_list) & set(leshy_rfc_list)
        ), "expect non-empty intersection"

    def test_borutaPy_vs_leshy_with_rfr_and_native_feature_importance(self):
        # too slow for circleci to run them in a reasonable time
        # takes 2 min on laptop, 1h or more on circleci
        # # sklearn random forest implementation
        # X, y, w = _generated_corr_dataset_regr()
        # rfr = RandomForestRegressor(max_features=0.3, max_samples=0.632, n_estimators=100)
        # bt = BorutaPy(rfr)
        # bt.fit(X.values, y)
        # borutapy_rfc_list = sorted(list(X.columns[bt.support_]))

        # lightGBM random forest implementation
        baseline_list = ["var0", "var1", "var2", "var3", "var4"]
        X, y, w = generated_corr_dataset_regr(size=100)
        n_feat = X.shape[1]
        rfr = LightForestRegressor(n_feat)
        # rfr = RandomForestRegressor(max_features=0.3, max_samples=0.632, n_estimators=10)
        arfs = Leshy(rfr, verbose=0, max_iter=10, random_state=42, importance="native")
        arfs.fit(X, y)
        leshy_rfc_list = sorted(arfs.support_names_)

        # assert borutapy_rfc_list == leshy_rfc_list, "same selected features are expected"
        assert bool(
            set(baseline_list) & set(leshy_rfc_list)
        ), "expect non-empty intersection"

    def test_borutaPy_vs_leshy_with_rfc_and_shap_feature_importance(self):
        # too slow for circleci to run them in a reasonable time
        # takes 2 min on laptop, 1h or more on circleci
        # # sklearn random forest implementation
        # X, y, w = _generated_corr_dataset_classification()
        # rfc = RandomForestClassifier(max_features='sqrt', max_samples=0.632, n_estimators=100)
        # bt = BorutaPy(rfc)
        # bt.fit(X.values, y)
        # borutapy_rfc_list = sorted(list(X.columns[bt.support_]))

        # lightGBM random forest implementation
        baseline_list = ["var0", "var1", "var2", "var3", "var4"]
        X, y, w = generated_corr_dataset_classification(size=100)
        n_feat = X.shape[1]
        model = LightForestClassifier(n_feat)
        arfs = Leshy(model, verbose=0, max_iter=10, random_state=42, importance="shap")
        arfs.fit(X, y)
        leshy_rfc_list = sorted(arfs.support_names_)

        # assert borutapy_rfc_list == leshy_rfc_list, "same selected features are expected"
        assert bool(
            set(baseline_list) & set(leshy_rfc_list)
        ), "expect non-empty intersection"

    def test_borutaPy_vs_leshy_with_rfr_and_shap_feature_importance(self):
        # too slow for circleci to run them in a reasonable time
        # takes 2 min on laptop, 1h or more on circleci
        # # sklearn random forest implementation
        # X, y, w = _generated_corr_dataset_regr()
        # rfr = RandomForestRegressor(max_features=0.3, max_samples=0.632, n_estimators=100)
        # bt = BorutaPy(rfr)
        # bt.fit(X.values, y)
        # borutapy_rfc_list = sorted(list(X.columns[bt.support_]))

        # lightGBM random forest implementation
        baseline_list = ["var0", "var1", "var2", "var3", "var4"]
        X, y, w = generated_corr_dataset_regr(size=500)
        n_feat = X.shape[1]
        model = LightForestRegressor(n_feat)
        arfs = Leshy(model, verbose=0, max_iter=10, random_state=42, importance="shap")
        arfs.fit(X, y)
        leshy_rfc_list = sorted(arfs.support_names_)

        # assert borutapy_rfc_list == leshy_rfc_list, "same selected features are expected"
        assert bool(
            set(baseline_list) & set(leshy_rfc_list)
        ), "expect non-empty intersection"

    def test_leshy_clf_with_lgb_and_shap_feature_importance_and_sample_weight(self):
        baseline_list = ["var0", "var1", "var2", "var3", "var4"]

        X, y, w = generated_corr_dataset_classification(size=500)
        model = lgb.LGBMClassifier(verbose=-1, force_col_wise=True, n_estimators=10)
        arfs = Leshy(model, verbose=0, max_iter=10, random_state=42, importance="shap")
        arfs.fit(X, y, w)
        leshy_list = sorted(arfs.support_names_)

        assert bool(
            set(baseline_list) & set(leshy_list)
        ), "expect non-empty intersection"

    def test_leshy_regr_with_lgb_and_shap_feature_importance_and_sample_weight(self):
        baseline_list = ["var0", "var1", "var2", "var3", "var4", "var5"]

        X, y, w = generated_corr_dataset_classification(size=500)
        model = lgb.LGBMRegressor(verbose=-1, force_col_wise=True, n_estimators=10)
        arfs = Leshy(model, verbose=0, max_iter=10, random_state=42, importance="shap")
        arfs.fit(X, y, w)
        leshy_list = sorted(arfs.support_names_)

        assert bool(
            set(baseline_list) & set(leshy_list)
        ), "expect non-empty intersection"


class TestBoostAGroota:
    """
    Test suite for all-relevant FS boruta-like method: Leshy
    """

    def test_boostagroota_clf_with_lgb_and_shap_feature_importance_and_sample_weight(
        self,
    ):
        baseline_list = ["var0", "var1", "var2", "var3", "var4"]

        X, y, w = generated_corr_dataset_classification(size=500)
        model = lgb.LGBMClassifier(verbose=-1, force_col_wise=True, n_estimators=10)
        arfs = BoostAGroota(
            est=model,
            cutoff=1,
            iters=3,
            max_rounds=3,
            delta=0.1,
            silent=False,
            importance="shap",
        )
        arfs.fit(X, y, w)
        leshy_list = sorted(arfs.support_names_)

        assert bool(
            set(baseline_list) & set(leshy_list)
        ), "expect non-empty intersection"

    def test_boostagroota_clf_with_lgb_and_pimp_feature_importance_and_sample_weight(
        self,
    ):
        baseline_list = ["var0", "var1", "var2", "var3", "var4"]

        X, y, w = generated_corr_dataset_classification(size=500)
        model = lgb.LGBMClassifier(verbose=-1, force_col_wise=True, n_estimators=10)
        arfs = BoostAGroota(
            est=model,
            cutoff=1,
            iters=3,
            max_rounds=3,
            delta=0.1,
            silent=False,
            importance="pimp",
        )
        arfs.fit(X, y, w)
        leshy_list = sorted(arfs.support_names_)

        assert bool(
            set(baseline_list) & set(leshy_list)
        ), "expect non-empty intersection"

    def test_boostagroota_rgr_with_lgb_and_shap_feature_importance_and_sample_weight(
        self,
    ):
        baseline_list = ["var0", "var1", "var2", "var3", "var4", "var5"]

        X, y, w = generated_corr_dataset_regr(size=500)
        model = lgb.LGBMRegressor(verbose=-1, force_col_wise=True, n_estimators=10)
        arfs = BoostAGroota(
            est=model,
            cutoff=1,
            iters=3,
            max_rounds=3,
            delta=0.1,
            silent=False,
            importance="shap",
        )
        arfs.fit(X, y, w)
        leshy_list = sorted(arfs.support_names_)

        assert bool(
            set(baseline_list) & set(leshy_list)
        ), "expect non-empty intersection"

    def test_boostagroota_regr_with_lgb_and_pimp_feature_importance_and_sample_weight(
        self,
    ):
        baseline_list = ["var0", "var1", "var2", "var3", "var4", "var5"]

        X, y, w = generated_corr_dataset_regr(size=500)
        model = lgb.LGBMRegressor(verbose=-1, force_col_wise=True, n_estimators=10)
        arfs = BoostAGroota(
            est=model,
            cutoff=1,
            iters=3,
            max_rounds=3,
            delta=0.1,
            silent=False,
            importance="pimp",
        )
        arfs.fit(X, y, w)
        leshy_list = sorted(arfs.support_names_)

        assert bool(
            set(baseline_list) & set(leshy_list)
        ), "expect non-empty intersection"


class TestGrootCV:
    """
    Test suite for all-relevant FS boruta-like method: Leshy
    """

    def test_grootcv_classification_with_and_sample_weight(self):
        baseline_list = ["var0", "var1", "var2", "var3", "var4"]

        X, y, w = generated_corr_dataset_classification(size=100)
        arfs = GrootCV(objective="binary", cutoff=1, n_folds=3, n_iter=3, silent=False)
        arfs.fit(X, y, w)
        grootcv_list = sorted(arfs.support_names_)

        assert bool(
            set(baseline_list) & set(grootcv_list)
        ), "expect non-empty intersection"

    def test_grootcv_regression_with_and_sample_weight(self):
        baseline_list = ["var0", "var1", "var2", "var3", "var4", "var5"]

        X, y, w = generated_corr_dataset_regr(size=100)
        arfs = GrootCV(objective="l2", cutoff=1, n_folds=3, n_iter=3, silent=False)
        arfs.fit(X, y, w)
        grootcv_list = sorted(arfs.support_names_)

        assert bool(
            set(baseline_list) & set(grootcv_list)
        ), "expect non-empty intersection"



import pytest
import numpy as np
import pandas as pd
from arfs.feature_selection import (
    MissingValueThreshold,
    UniqueValuesThreshold,
    CardinalityThreshold,
    CollinearityThreshold,
)
from arfs.utils import (
    _generated_corr_dataset_regr,
    _generated_corr_dataset_classification,
)


class TestFeatSelectMissing:
    """
    Test suite for FeatureSelector, missing values
    """

    def test_identify_missing_for_classification(self):
        # not task dependent (same for clf and regr)
        X, y, w = _generated_corr_dataset_classification(size=10)
        fs = MissingValueThreshold(threshold=0.01)
        fs.fit(X)
        message = "Expected: {0}, Actual: {1}".format(
            "var12", fs.not_selected_features_
        )
        assert fs.not_selected_features_ == ["var12"], message


class TestFeatSelectZeroVariance:
    """
    Test suite for FeatureSelector, missing values
    """

    def test_identify_single_unique_classification(self):
        # not task dependent (same for clf and regr)
        X, y, w = _generated_corr_dataset_classification(size=10)
        fs = UniqueValuesThreshold(threshold=1)
        fs.fit(X)
        message = "Expected: {0}, Actual: {1}".format(
            "var10", fs.not_selected_features_
        )
        assert fs.not_selected_features_ == ["var10"], message


class TestFeatSelectHighCardinality:
    """
    Test suite for FeatureSelector, high cardinality
    """

    def test_identify_high_cardinality_classification(self):
        # not task dependent (same for clf and regr)
        X, y, w = _generated_corr_dataset_classification(size=100)
        fs = CardinalityThreshold(threshold=5)
        fs.fit(X)
        message = "Expected: {0}, Actual: {1}".format(
            "emb_dummy", fs.not_selected_features_
        )
        assert fs.not_selected_features_ == ["emb_dummy"], message


# class TestFeatSelectCollinearity:
#     """
#     test suite for FeatureSelector, high cardinality
#     """

#     def test_identify_collinear_spearman_no_encoding(self):
#         X, y, w = _generated_corr_dataset_regr(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         fs.identify_collinear(correlation_threshold=0.5, encode=False, method='spearman')
#         message = "Expected: {0}, Actual: {1}".format(['var2', 'var3', 'var4', 'var12'], fs.ops['collinear'])
#         assert fs.ops['collinear'] == ['var2', 'var3', 'var4', 'var12'], message

#     def test_identify_collinear_pearson_no_encoding(self):
#         X, y, w = _generated_corr_dataset_regr(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         fs.identify_collinear(correlation_threshold=0.5, encode=False, method='pearson')
#         message = "Expected: {0}, Actual: {1}".format(['var2', 'var3', 'var12'], fs.ops['collinear'])
#         assert fs.ops['collinear'] == ['var2', 'var3', 'var12'], message

#     def test_identify_collinear_spearman_with_encoding(self):
#         X, y, w = _generated_corr_dataset_regr(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         fs.identify_collinear(correlation_threshold=0.5, encode=True, method='spearman')
#         message = "Expected: {0}, Actual: {1}".format(['var2', 'var3', 'var4', 'var12'], fs.ops['collinear'])
#         assert fs.ops['collinear'] == ['var2', 'var3', 'var4', 'var12'], message

#     def test_identify_collinear_pearson_with_encoding(self):
#         X, y, w = _generated_corr_dataset_regr(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         fs.identify_collinear(correlation_threshold=0.5, encode=True, method='pearson')
#         message = "Expected: {0}, Actual: {1}".format(['var2', 'var3', 'var12'], fs.ops['collinear'])
#         assert fs.ops['collinear'] == ['var2', 'var3', 'var12'], message


# class TestFeatSelectZeroImportance:
#     """
#     test suite for FeatureSelector, high cardinality
#     """

#     def test_identify_zero_importance_for_regression_with_early_stopping(self):
#         X, y, w = _generated_corr_dataset_regr(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         fs.identify_zero_importance(task='regression', eval_metric='l2', objective='l2', n_iterations=2,
#                                     early_stopping=True)
#         message = "Expected: {0}, Actual: {1}".format(['var10'], fs.ops['zero_importance'])
#         assert 'var10' in fs.ops['zero_importance'], message

#     @pytest.mark.xfail
#     def test_identify_zero_importance_for_regression_with_early_stopping_no_eval_metric(self):
#         X, y, w = _generated_corr_dataset_regr(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         # Xfail: expected to fail because the eval metric is not provided
#         fs.identify_zero_importance(task='regression', eval_metric=None, objective='l2', n_iterations=2,
#                                     early_stopping=True)
#         message = "Expected: {0}, Actual: {1}".format(['var10'], fs.ops['zero_importance'])
#         assert 'var10' in fs.ops['zero_importance'], message

#     @pytest.mark.xfail
#     def test_identify_zero_importance_for_regression_with_early_stopping_no_eval_metric_no_objective(self):
#         X, y, w = _generated_corr_dataset_regr(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         # Xfail: expected to fail because the eval metric is not provided
#         fs.identify_zero_importance(task='regression', eval_metric=None, objective=None, n_iterations=2,
#                                     early_stopping=True)
#         message = "Expected: {0}, Actual: {1}".format(['var10'], fs.ops['zero_importance'])
#         assert 'var10' in fs.ops['zero_importance'], message

#     @pytest.mark.xfail
#     def test_identify_zero_importance_for_regression_with_early_stopping_wrong_task(self):
#         X, y, w = _generated_corr_dataset_regr(size=10)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         # Xfail: expected to fail because the eval metric is not provided
#         fs.identify_zero_importance(task='classification', eval_metric='l2', objective='l2', n_iterations=2,
#                                     early_stopping=True)
#         message = "Expected: {0}, Actual: {1}".format(['var10'], fs.ops['zero_importance'])
#         assert 'var10' in fs.ops['zero_importance'], message

#     def test_identify_zero_importance_for_regression_without_early_stopping(self):
#         X, y, w = _generated_corr_dataset_regr(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         fs.identify_zero_importance(task='regression', eval_metric='l2', objective='l2', n_iterations=2,
#                                     early_stopping=False)
#         message = "Expected: {0}, Actual: {1}".format(['var10'], fs.ops['zero_importance'])
#         assert 'var10' in fs.ops['zero_importance'], message

#     def test_identify_zero_importance_for_regression_without_early_stopping_no_objective(self):
#         X, y, w = _generated_corr_dataset_regr(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         fs.identify_zero_importance(task='regression', n_iterations=2,
#                                     early_stopping=False)
#         message = "Expected: {0}, Actual: {1}".format(['var10'], fs.ops['zero_importance'])
#         assert 'var10' in fs.ops['zero_importance'], message

#     def test_identify_zero_importance_for_classification_with_early_stopping(self):
#         X, y, w = _generated_corr_dataset_classification(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         fs.identify_zero_importance(task='classification', eval_metric='auc', n_iterations=2,
#                                     early_stopping=True)
#         message = "Expected: {0}, Actual: {1}".format(['var10'], fs.ops['zero_importance'])
#         assert 'var10' in fs.ops['zero_importance'], message

#     @pytest.mark.xfail
#     def test_identify_zero_importance_for_classification_with_early_stopping_no_eval_metric(self):
#         X, y, w = _generated_corr_dataset_classification(size=10)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         # Xfail: expected to fail because the eval metric is not provided
#         fs.identify_zero_importance(task='classification', eval_metric=None, n_iterations=2,
#                                     early_stopping=True)
#         message = "Expected: {0}, Actual: {1}".format(['var10'], fs.ops['zero_importance'])
#         assert 'var10' in fs.ops['zero_importance'], message

#     @pytest.mark.xfail
#     def test_identify_zero_importance_for_classification_with_early_stopping_no_eval_metric_no_objective(self):
#         X, y, w = _generated_corr_dataset_classification(size=10)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         # Xfail: expected to fail because the eval metric is not provided
#         fs.identify_zero_importance(task='classification', eval_metric=None, objective=None, n_iterations=2,
#                                     early_stopping=True)
#         message = "Expected: {0}, Actual: {1}".format(['var10'], fs.ops['zero_importance'])
#         assert 'var10' in fs.ops['zero_importance'], message

#     @pytest.mark.xfail
#     def test_identify_zero_importance_for_classification_with_early_stopping_wrong_task(self):
#         X, y, w = _generated_corr_dataset_classification(size=10)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         # Xfail: expected to fail because the eval metric is not provided
#         fs.identify_zero_importance(task='regression', eval_metric='auc', objective='cross-entropy', n_iterations=2,
#                                     early_stopping=True)
#         message = "Expected: {0}, Actual: {1}".format(['var10'], fs.ops['zero_importance'])
#         assert 'var10' in fs.ops['zero_importance'], message

#     def test_identify_zero_importance_for_classification_without_early_stopping(self):
#         X, y, w = _generated_corr_dataset_classification(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         fs.identify_zero_importance(task='classification', objective='binary', n_iterations=2,
#                                     early_stopping=False)
#         message = "Expected: {0}, Actual: {1}".format(['var10'], fs.ops['zero_importance'])
#         assert 'var10' in fs.ops['zero_importance'], message

#     def test_identify_zero_importance_for_classification_without_early_stopping_no_objective(self):
#         X, y, w = _generated_corr_dataset_classification(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         fs.identify_zero_importance(task='classification', n_iterations=2,
#                                     early_stopping=False)
#         message = "Expected: {0}, Actual: {1}".format(['var10'], fs.ops['zero_importance'])
#         assert 'var10' in fs.ops['zero_importance'], message


# class TestFeatSelectLowImportance:
#     """
#     test suite for FeatureSelector, high cardinality
#     """

#     def test_identify_low_importance_for_regression_with_early_stopping(self):
#         X, y, w = _generated_corr_dataset_regr(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         fs.identify_zero_importance(task='regression', eval_metric='l2', objective='l2', n_iterations=2,
#                                     early_stopping=True)
#         cum_imp_threshold = 0.95
#         fs.identify_low_importance(cumulative_importance=cum_imp_threshold)
#         expected = 1
#         message = "Expected at least one predictor ruled out, Actual: {0}".format(sorted(fs.ops['low_importance']))
#         assert len(fs.ops['low_importance']) >= expected, message

#     @pytest.mark.xfail
#     def test_identify_low_importance_for_regression_with_early_stopping_no_eval_metric(self):
#         X, y, w = _generated_corr_dataset_regr(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         # Xfail: expected to fail because the eval metric is not provided
#         fs.identify_zero_importance(task='regression', eval_metric=None, objective='l2', n_iterations=2,
#                                     early_stopping=True)
#         cum_imp_threshold = 0.95
#         fs.identify_low_importance(cumulative_importance=cum_imp_threshold)
#         expected = 1
#         message = "Expected at least one predictor ruled out, Actual: {0}".format(sorted(fs.ops['low_importance']))
#         assert len(fs.ops['low_importance']) >= expected, message

#     @pytest.mark.xfail
#     def test_identify_low_importance_for_regression_with_early_stopping_no_eval_metric_no_objective(self):
#         X, y, w = _generated_corr_dataset_regr(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         # Xfail: expected to fail because the eval metric is not provided
#         fs.identify_zero_importance(task='regression', eval_metric=None, objective=None, n_iterations=2,
#                                     early_stopping=True)
#         cum_imp_threshold = 0.95
#         fs.identify_low_importance(cumulative_importance=cum_imp_threshold)
#         expected = 1
#         message = "Expected at least one predictor ruled out, Actual: {0}".format(sorted(fs.ops['low_importance']))
#         assert len(fs.ops['low_importance']) >= expected, message

#     @pytest.mark.xfail
#     def test_identify_low_importance_for_regression_with_early_stopping_wrong_task(self):
#         X, y, w = _generated_corr_dataset_regr(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         # Xfail: expected to fail because the eval metric is not provided
#         fs.identify_zero_importance(task='classification', eval_metric='l2', objective='l2', n_iterations=2,
#                                     early_stopping=True)
#         cum_imp_threshold = 0.95
#         fs.identify_low_importance(cumulative_importance=cum_imp_threshold)
#         expected = 1
#         message = "Expected at least one predictor ruled out, Actual: {0}".format(sorted(fs.ops['low_importance']))
#         assert len(fs.ops['low_importance']) >= expected, message

#     def test_identify_low_importance_for_regression_without_early_stopping(self):
#         X, y, w = _generated_corr_dataset_regr(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         fs.identify_zero_importance(task='regression', objective='l2', n_iterations=2, early_stopping=False)
#         cum_imp_threshold = 0.95
#         fs.identify_low_importance(cumulative_importance=cum_imp_threshold)
#         expected = 1
#         message = "Expected at least one predictor ruled out, Actual: {0}".format(sorted(fs.ops['low_importance']))
#         assert len(fs.ops['low_importance']) >= expected, message

#     def test_identify_low_importance_for_regression_without_early_stopping_no_objective(self):
#         X, y, w = _generated_corr_dataset_regr(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         fs.identify_zero_importance(task='regression', n_iterations=2, early_stopping=False)
#         cum_imp_threshold = 0.95
#         fs.identify_low_importance(cumulative_importance=cum_imp_threshold)
#         expected = 1
#         message = "Expected at least one predictor ruled out, Actual: {0}".format(sorted(fs.ops['low_importance']))
#         assert len(fs.ops['low_importance']) >= expected, message

#     def test_identify_low_importance_for_classification_with_early_stopping(self):
#         X, y, w = _generated_corr_dataset_classification(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         fs.identify_zero_importance(task='classification', eval_metric='auc', n_iterations=2, early_stopping=True)
#         cum_imp_threshold = 0.95
#         fs.identify_low_importance(cumulative_importance=cum_imp_threshold)
#         expected = 1
#         message = "Expected at least one predictor ruled out, Actual: {0}".format(sorted(fs.ops['low_importance']))
#         assert len(fs.ops['low_importance']) >= expected, message

#     @pytest.mark.xfail
#     def test_identify_low_importance_for_classification_with_early_stopping_no_eval_metric(self):
#         X, y, w = _generated_corr_dataset_classification(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         # Xfail: expected to fail because the eval metric is not provided
#         fs.identify_zero_importance(task='classification', eval_metric=None, n_iterations=2,
#                                     early_stopping=True)
#         cum_imp_threshold = 0.95
#         fs.identify_low_importance(cumulative_importance=cum_imp_threshold)
#         expected = 1
#         message = "Expected at least one predictor ruled out, Actual: {0}".format(sorted(fs.ops['low_importance']))
#         assert len(fs.ops['low_importance']) >= expected, message

#     @pytest.mark.xfail
#     def test_identify_low_importance_for_classification_with_early_stopping_no_eval_metric_no_objective(self):
#         X, y, w = _generated_corr_dataset_classification(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         # Xfail: expected to fail because the eval metric is not provided
#         fs.identify_zero_importance(task='classification', eval_metric=None, objective=None, n_iterations=2,
#                                     early_stopping=True)
#         cum_imp_threshold = 0.95
#         fs.identify_low_importance(cumulative_importance=cum_imp_threshold)
#         expected = 1
#         message = "Expected at least one predictor ruled out, Actual: {0}".format(sorted(fs.ops['low_importance']))
#         assert len(fs.ops['low_importance']) >= expected, message

#     @pytest.mark.xfail
#     def test_identify_low_importance_for_classification_with_early_stopping_wrong_task(self):
#         X, y, w = _generated_corr_dataset_classification(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         # Xfail: expected to fail because the eval metric is not provided
#         fs.identify_zero_importance(task='regression', eval_metric='auc', objective='cross-entropy', n_iterations=2,
#                                     early_stopping=True)
#         cum_imp_threshold = 0.95
#         fs.identify_low_importance(cumulative_importance=cum_imp_threshold)
#         expected = 1
#         message = "Expected at least one predictor ruled out, Actual: {0}".format(sorted(fs.ops['low_importance']))
#         assert len(fs.ops['low_importance']) >= expected, message

#     def test_identify_low_importance_for_classification_without_early_stopping(self):
#         X, y, w = _generated_corr_dataset_classification(size=100)
#         fs = FeatureSelector(X=X, y=y, sample_weight=w)
#         fs.identify_zero_importance(task='classification', n_iterations=2, early_stopping=False)
#         cum_imp_threshold = 0.95
#         fs.identify_low_importance(cumulative_importance=cum_imp_threshold)
#         expected = 1
#         message = "Expected at least one predictor ruled out, Actual: {0}".format(sorted(fs.ops['low_importance']))
#         assert len(fs.ops['low_importance']) >= expected, message



"""
This module provides a function for creating the summary report of a FS pipeline

**The module structure is the following:**

- The ``make_fs_summary`` main function for creating the summary
- The ``highlight_discarded`` function for creating style for the pd.DataFrame

"""

import pandas as pd
import numpy as np


def highlight_discarded(s):
    """highlight X in red and V in green.

    Parameters
    ----------
    s : array-like of shape (n_features,)
        the boolean array for defining the style


    """
    is_X = s == 0
    return [
        "background-color: #d65f5f" if v else "background-color: #33a654" for v in is_X
    ]


def make_fs_summary(selector_pipe):
    """make_fs_summary makes a summary dataframe highlighting at which step a
    given predictor has been rejected (if any).

    Parameters
    ----------
    selector_pipe : sklearn.pipeline.Pipeline
        the feature selector pipeline.

    Examples
    --------
    >>> groot_pipeline = Pipeline([
    ... ('missing', MissingValueThreshold()),
    ... ('unique', UniqueValuesThreshold()),
    ... ('cardinality', CardinalityThreshold()),
    ... ('collinearity', CollinearityThreshold(threshold=0.5)),
    ... ('lowimp', VariableImportance(eval_metric='poisson', objective='poisson', verbose=2)),
    ... ('grootcv', GrootCV(objective='poisson', cutoff=1, n_folds=3, n_iter=5))])
    >>> groot_pipeline.fit_transform(
        X=df[predictors],
        y=df[target],
        lowimp__sample_weight=df[weight],
        grootcv__sample_weight=df[weight])
    >>> fs_summary_df = make_fs_summary(groot_pipeline)
    """

    tag_df = pd.DataFrame({"predictor": selector_pipe[0].feature_names_in_})
    for selector_name in selector_pipe.named_steps.keys():
        if hasattr(selector_pipe.named_steps[selector_name], "support_"):
            feature_in = selector_pipe.named_steps[selector_name].feature_names_in_
            to_drop = list(
                set(selector_pipe.named_steps[selector_name].feature_names_in_)
                - set(selector_pipe.named_steps[selector_name].get_feature_names_out())
            )
            tag_df[selector_name] = np.where(
                tag_df["predictor"].isin(to_drop), 0, 1
            ) * np.where(tag_df["predictor"].isin(feature_in), 1, np.nan)

    col_to_apply_style = tag_df.columns[1:]
    tag_df = (
        tag_df.style.apply(highlight_discarded, subset=col_to_apply_style)
        .applymap(lambda x: "" if x == x else "background-color: #ffa500")
        .format(precision=0)
    )
    return tag_df



"""
This module provides preprocessing classes

**The module structure is the following:**

- The ``OrdinalEncoderPandas`` main class for ordinal encoding, takes in a DF and returns a DF of the same shape
- The ``dtype_column_selector`` for standardizing selection of columns based on their dtypes
- The ``TreeDiscretizer`` for discretizing continuous columns and auto-group levels of categorical columns

"""

# Settings and libraries
from __future__ import print_function
from tqdm import tqdm

# pandas
import pandas as pd

# numpy
import numpy as np

# sklearn
from sklearn.preprocessing import OrdinalEncoder
from sklearn.base import BaseEstimator, TransformerMixin

# ARFS
from .gbm import GradientBoosting


# fix random seed for reproducibility
np.random.seed(7)

__all__ = [
    "OrdinalEncoderPandas",
    "dtype_column_selector",
    "TreeDiscretizer",
]


class OrdinalEncoderPandas(OrdinalEncoder):
    # class OrdinalEncoderPandas(BaseEstimator, TransformerMixin):
    """Encode categorical features as an integer array and returns a pandas DF.
    The features are converted to ordinal integers. This results in
    a single column of integers (0 to n_categories - 1) per feature.
    Read more in the scikit-learn OrdinalEncoder documentation

    Parameters
    ----------
    pattern : str, default=None
        Name of columns containing this regex pattern will be included. If
        None, column selection will not be selected based on pattern.
    dtype_include : column dtype or list of column dtypes, default=None
        A selection of dtypes to include. For more details, see
        `pandas.DataFrame.select_dtypes`.
    dtype_exclude : column dtype or list of column dtypes, default=None
        A selection of dtypes to exclude. For more details, see
        `pandas.DataFrame.select_dtypes`.
    exclude_cols : list of str, optional
        columns to not encode
    output_dtype : number type, default np.float64
        Desired dtype of output.
    handle_unknown : {'error', 'use_encoded_value'}, default='error'
        When set to 'error' an error will be raised in case an unknown
        categorical feature is present during transform. When set to
        'use_encoded_value', the encoded value of unknown categories will be
        set to the value given for the parameter `unknown_value`. In
        `inverse_transform`, an unknown category will be denoted as None.
    unknown_value : int or np.nan, default=None
        When the parameter handle_unknown is set to 'use_encoded_value', this
        parameter is required and will set the encoded value of unknown
        categories. It has to be distinct from the values used to encode any of
        the categories in `fit`. If set to np.nan, the `dtype` parameter must
        be a float dtype.
    encoded_missing_value : int or np.nan, default=np.nan
        Encoded value of missing categories. If set to `np.nan`, then the `dtype`
        parameter must be a float dtype.
    return_pandas_categorical : bool, defult=False
        return encoded columns as pandas category dtype or as float

    Attributes
    ----------
    categories_ : list of arrays
        The categories of each feature determined during ``fit`` (in order of
        the features in X and corresponding with the output of ``transform``).
        This does not include categories that weren't seen during ``fit``.
    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Defined only when `X`
        has feature names that are all strings.

    Examples
    --------
    Given a dataset with two features, we let the encoder find the unique
    values per feature and transform the data to an ordinal encoding.
    >>> ord_enc = OrdinalEncoderPandas(exclude_cols=["PARENT1", "SEX"])
    >>> X_enc = ord_enc.fit_transform(X)
    >>> X_original = ord_enc.inverse_transform(X_enc)
    """

    def __init__(
        self,
        dtype_include=["category", "object", "bool"],
        dtype_exclude=[np.number],
        pattern=None,
        exclude_cols=None,
        output_dtype=np.float64,
        handle_unknown="use_encoded_value",
        unknown_value=np.nan,
        encoded_missing_value=np.nan,
        return_pandas_categorical=False,
    ):

        self.dtype_include = dtype_include
        self.dtype_exclude = dtype_exclude
        self.pattern = pattern
        self.exclude_cols = exclude_cols
        self.output_dtype = output_dtype
        self.handle_unknown = handle_unknown
        self.unknown_value = unknown_value
        self.encoded_missing_value = encoded_missing_value
        self.return_pandas_categorical = return_pandas_categorical

        super().__init__(
            categories="auto",
            dtype=self.output_dtype,
            handle_unknown=self.handle_unknown,
            unknown_value=self.unknown_value,
            encoded_missing_value=self.encoded_missing_value,
        )

    def fit(self, X, y=None):
        """
        Fit the OrdinalEncoder to X.

        Parameters
        ----------
        X : pd.DataFrame, of shape (n_samples, n_features)
            The data to determine the categories of each feature.
        y : Ignored. This parameter exists only for compatibility with
            :class:`~sklearn.pipeline.Pipeline`.

        Returns
        -------
        self :
            Fitted encoder.
        """

        cat_features_selector = dtype_column_selector(
            dtype_include=self.dtype_include,
            dtype_exclude=self.dtype_exclude,
            pattern=self.pattern,
            exclude_cols=self.exclude_cols,
        )

        self.feature_names_in_ = X.columns.to_numpy()
        self.categorical_features_ = cat_features_selector(X)

        super(OrdinalEncoderPandas, self).fit(X[self.categorical_features_])
        # self.feature_names_in_ = X.columns.to_numpy()
        return self

    def transform(self, X, y=None, sample_weight=None):
        """
        Transform X to ordinal codes.

        Parameters
        ----------
        X : pd.DataFrame of shape (n_samples, n_features)
            The data to encode.

        Returns
        -------
        X_out : pd.DataFrame (n_samples, n_features)
            Transformed input.
        """
        X_trans = X.copy()
        X_trans[self.categorical_features_] = super(
            OrdinalEncoderPandas, self
        ).transform(X_trans[self.categorical_features_])
        # X_trans.loc[:, self.feature_names_in_] = X_trans.loc[:, self.feature_names_in_].apply(pd.to_numeric, errors="ignore", axis=1)

        if self.return_pandas_categorical:
            X_trans[self.categorical_features_] = X_trans[
                self.categorical_features_
            ].astype("category")
        return X_trans

    def fit_transform(self, X, y=None, sample_weight=None, **fit_params):
        """
        Fit to data, then transform it.
        Fits transformer to `X` and `y` with optional parameters `fit_params`
        and returns a transformed version of `X`.
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Input samples.
        y :  array-like of shape (n_samples,) or (n_samples, n_outputs), \
                default=None
            Target values (None for unsupervised transformations).
        **fit_params : dict
            Additional fit parameters.
        Returns
        -------
        X_new : ndarray array of shape (n_samples, n_features_new)
            Transformed array.
        """
        self = self.fit(X)
        return self.transform(X)

    def inverse_transform(self, X):
        """
        Convert the data back to the original representation.
        When unknown categories are encountered (all zeros in the
        one-hot encoding), ``None`` is used to represent this category. If the
        feature with the unknown category has a dropped category, the dropped
        category will be its inverse.
        For a given input feature, if there is an infrequent category,
        'infrequent_sklearn' will be used to represent the infrequent category.

        Parameters
        ----------
        X : pd.DataFrame of shape (n_samples, n_encoded_features)
            The transformed data.
        Returns
        -------
        X_tr : pd.Dataframe of shape (n_samples, n_features)
            Inverse transformed array.
        """

        X[self.categorical_features_] = super(
            OrdinalEncoderPandas, self
        ).inverse_transform(X[self.categorical_features_])
        return X


class dtype_column_selector:
    """Create a callable to select columns to be used with
    :class:`ColumnTransformer`.
    :func:`dtype_column_selector` can select columns based on datatype or the
    columns name with a regex. When using multiple selection criteria, **all**
    criteria must match for a column to be selected.

    Parameters
    ----------
    pattern : str, default=None
        Name of columns containing this regex pattern will be included. If
        None, column selection will not be selected based on pattern.
    dtype_include : column dtype or list of column dtypes, default=None
        A selection of dtypes to include. For more details, see
        :meth:`pandas.DataFrame.select_dtypes`.
    dtype_exclude : column dtype or list of column dtypes, default=None
        A selection of dtypes to exclude. For more details, see
        :meth:`pandas.DataFrame.select_dtypes`.
    exclude_cols : list of column names, default=None
        A selection of columns to exclude

    Returns
    -------
    selector : callable
        Callable for column selection to be used by a
        :class:`ColumnTransformer`.

    See Also
    --------
    ColumnTransformer : Class that allows combining the
        outputs of multiple transformer objects used on column subsets
        of the data into a single feature space.

    Examples
    --------
    >>> from sklearn.preprocessing import StandardScaler, OneHotEncoder
    >>> from sklearn.compose import make_column_transformer
    >>> from arfs.preprocessing import dtype_column_selector
    >>> import numpy as np
    >>> import pandas as pd  # doctest: +SKIP
    >>> X = pd.DataFrame({'city': ['London', 'London', 'Paris', 'Sallisaw'],
    ...                   'rating': [5, 3, 4, 5]})  # doctest: +SKIP
    >>> ct = make_column_transformer(
    ...       (StandardScaler(),
    ...        dtype_column_selector(dtype_include=np.number)),  # rating
    ...       (OneHotEncoder(),
    ...        dtype_column_selector(dtype_include=object)))  # city
    >>> ct.fit_transform(X)
    array([[ 0.90453403,  1.        ,  0.        ,  0.        ],
           [-1.50755672,  1.        ,  0.        ,  0.        ],
           [-0.30151134,  0.        ,  1.        ,  0.        ],
           [ 0.90453403,  0.        ,  0.        ,  1.        ]])
    """

    def __init__(
        self, pattern=None, *, dtype_include=None, dtype_exclude=None, exclude_cols=None
    ):
        self.pattern = pattern
        self.dtype_include = dtype_include
        self.dtype_exclude = dtype_exclude
        self.exclude_cols = exclude_cols

    def __call__(self, df):
        """Callable for column selection to be used by a
        :class:`ColumnTransformer`.
        Parameters
        ----------
        df : pd.DataFrame of shape (n_features, n_samples)
            DataFrame to select columns from.
        """
        if not hasattr(df, "iloc"):
            raise ValueError(
                "make_column_selector can only be applied to pandas dataframes"
            )
        df_row = df.iloc[:1]
        if self.dtype_include is not None or self.dtype_exclude is not None:
            df_row = df_row.select_dtypes(
                include=self.dtype_include, exclude=self.dtype_exclude
            )
        cols = df_row.columns
        if self.pattern is not None:
            cols = cols[cols.str.contains(self.pattern, regex=True)]

        if self.exclude_cols is not None:
            cols = cols[~cols.isin(self.exclude_cols)]

        return cols.tolist()


def cat_var(data, col_excl=None, return_cat=True):
    """Ad hoc categorical encoding (as integer). Automatically detect the non-numerical columns,
    save the index and name of those columns, encode them as integer,
    save the direct and inverse mappers as
    dictionaries.
    Return the data-set with the encoded columns with a data type either int or pandas categorical.

    Parameters
    ----------
    data: pd.DataFrame
        the dataset
    col_excl: list of str, default=None
        the list of columns names not being encoded (e.g. the ID column)
    return_cat: bool, default=True
        return encoded object columns as pandas categoricals or not.

    Returns
    -------
    df: pd.DataFrame
        the dataframe with encoded columns
    cat_var_df: pd.DataFrame
        the dataframe with the indices and names of the categorical columns
    inv_mapper: dict
        the dictionary to map integer --> category
    mapper: dict
        the dictionary to map category --> integer
    """
    df = data.copy()
    if col_excl is None:
        non_num_cols = list(
            set(list(df.columns)) - set(list(df.select_dtypes(include=[np.number])))
        )
    else:
        non_num_cols = list(
            set(list(df.columns))
            - set(list(df.select_dtypes(include=[np.number])))
            - set(col_excl)
        )
    cat_var_index = [df.columns.get_loc(c) for c in non_num_cols if c in df]
    cat_var_df = pd.DataFrame({"cat_ind": cat_var_index, "cat_name": non_num_cols})
    # avoid having datetime objects as keys in the mapping dic
    date_cols = [s for s in list(df) if "date" in s]
    df.loc[:, date_cols] = df.loc[:, date_cols].astype(str)
    cols_need_mapped = cat_var_df.cat_name.to_list()
    inv_mapper = {
        col: dict(enumerate(df[col].astype("category").cat.categories))
        for col in df[cols_need_mapped]
    }
    mapper = {
        col: {v: k for k, v in inv_mapper[col].items()} for col in df[cols_need_mapped]
    }
    progress_bar = tqdm(cols_need_mapped)
    for c in progress_bar:
        progress_bar.set_description("Processing {0:<30}".format(c))
        df.loc[:, c] = df.loc[:, c].map(mapper[c]).fillna(0).astype(int)
        # I could have use df[c].update(df[c].map(mapper[c])) while slower,
        # prevents values not included in an incomplete map from being changed to nans.
        # But then I could have outputs
        # with mixed types in the case of different dtypes mapping (like str -> int).
        # This would eventually break any flow.
        # Map is faster than replace
    if return_cat:
        df.loc[:, non_num_cols] = df.loc[:, non_num_cols].astype("category")
    return df, cat_var_df, inv_mapper, mapper


class TreeDiscretizer(BaseEstimator, TransformerMixin):
    """Bin continuous data into intervals and return a pandas DF. It supports regression and binary classification.
    Discretization (otherwise known as quantization or binning) provides a way to partition continuous features into discrete values.
    Certain datasets with continuous features may benefit from discretization, because discretization can transform the dataset
    of continuous attributes to one with only nominal attributes.
    One-hot encoded discretized features can make a model more expressive, while maintaining interpretability.
    For instance, pre-processing with a discretizer can introduce nonlinearity to linear models.
    For more advanced possibilities, in particular smooth ones, see Generating polynomial features further below.

    TreeDiscretizer uses univariate regularized trees (one per column to bin), find the optimal partition and returns
    pandas numerical interval for numerical continuous columns and pd.Categorical for categorical columns.
    Similar levels are therefore grouped together. This reduces the dimensionality and regularizes the model.
    This is a substitution to proper regularization scheme such as
     - GroupLasso (categorical predictors, which are usually encoded as multiple dummy variables: one for each category.
                   It makes sense in many analyses to consider these dummy variables (representing one categorical predictor)
                   together rather than separately)
     - FusedLasso (One drawback of the Lasso is it ignores ordering of the features, FusedLasso takes into account the ordering)

    TreeDiscretizer handles missing values, both for numerical and categorical predictors. No need for encoding the categorical predictors.


    Parameters
    ----------
    bin_features : List of string, str or None
        the list of names of the variable that has to be binned, or "all", "numerical" or "categorical"
        for splitting and grouping all, only numerical or only categorical columns.
    n_bins : int
        the number of bins that has to be created while binning the variables in "bin_features" list
    boost_params : dic
        the boosting parameters dictionary
    raw : bool
        returns raw levels (non human-interpretable) or levels matching the orginal ones
    task : str
        either regression or classification (binary)

    Attributes
    ----------
    tree_dic : dic
        the dictionary keys are binned column names and items are the univariate trees
    bin_upper_bound_dic : dic
        the upper bound of the numerical intervals
    cat_bin_dict : dic
        the mapping dictionary for the categorical columns
    tree_imputer : dic
        the missing values are split by the tree and lead to similar splits and are mapped to this value
    ordinal_encoder_dic : dic
        dictionary with the fitted encoder, if any
    cat_features : list
        names of the found categorical columns

    Methods
    -------
    fit(X, y, sample_weight=None)
        fit the transformer object on data
    transform(x)
        apply the fitted transformer object on new data
    fit_transform(x)
        fit and apply the transformer object on data

    Example
    -------
    >>> lgb_params = {'min_split_gain': 5}
    >>> disc = TreeDiscretizer(bin_features='all', n_bins=10)
    >>> disc.fit(X=df[predictors], y=df['Frequency'], sample_weight=df['Exposure'])
    """

    def __init__(
        self,
        bin_features="all",
        n_bins=10,
        boost_params=None,
        raw=False,
        task="regression",
    ):

        if (boost_params is not None) & (not isinstance(boost_params, dict)):
            raise TypeError("boost_kwargs should be a dictionary")

        self.bin_features = bin_features
        self.n_bins = n_bins
        self.boost_params = {}
        self.raw = raw
        self.task = task
        if boost_params is not None:
            self.boost_params = boost_params

        # force some params
        if self.task == "regression":
            self.boost_params["objective"] = "RMSE"
        elif self.task == "classification":
            self.boost_params["objective"] = "binary"

        self.boost_params["num_boost_round"] = 1
        self.boost_params["max_leaf"] = self.n_bins
        self.tree_dic = {}
        self.bin_upper_bound_dic = {}
        self.cat_bin_dict = {}
        self.tree_imputer = {}
        self.ordinal_encoder_dic = {}
        self.cat_features = None

    def fit(self, X, y, sample_weight=None):
        """Fit the discretizer on `X`.
        Parameters
        ----------
        X :
            Input data shape (n_samples, n_features), where `n_samples` is the number of samples and
            `n_features` is the number of features.
        y :
            target for internally fitting the tree(s)
        sample_weight :
            sample weight (e.g. exposure) if any

        Returns
        -------
        X :
            pd.DataFrame with the binned and grouped columns
        """
        X = X.copy()
        if self.bin_features is None:
            self.bin_features = list(X.select_dtypes("number").columns)
            self.cat_features = []
        elif isinstance(self.bin_features, list):
            self.cat_features = list(
                set(self.bin_features)
                - set(list(X[self.bin_features].select_dtypes("number").columns))
            )
        elif isinstance(self.bin_features, str) and (self.bin_features == "all"):
            self.bin_features = list(X.columns)
            self.cat_features = list(
                set(self.bin_features) - set(list(X.select_dtypes("number").columns))
            )
        elif isinstance(self.bin_features, str) and (self.bin_features == "numerical"):
            self.bin_features = list(X.select_dtypes("number").columns)
        elif isinstance(self.bin_features, str) and (
            self.bin_features == "categorical"
        ):
            self.bin_features = list(X.select_dtypes(["category", "object"]).columns)
            self.cat_features = self.bin_features

        for col in self.bin_features:
            if (self.cat_features is not None) and (col in self.cat_features):
                encoder = OrdinalEncoder(
                    handle_unknown="use_encoded_value", unknown_value=np.nan
                )
                # create a category for missing
                # X[col] = X[col].astype("category").cat.add_categories("missing").fillna("missing")
                X[col] = (
                    X[col]
                    .astype("category")
                    .cat.add_categories("missing_added")
                    .fillna("missing_added")
                )
                # encode
                self.ordinal_encoder_dic[col] = encoder.fit(X[[col]])
                X[col] = encoder.transform(X[[col]]).ravel()

            gbm_param = self.boost_params.copy()
            tree = GradientBoosting(
                cat_feat=None,
                params=gbm_param,
                show_learning_curve=False,
            )

            tree.fit(X[[col]], y, sample_weight=sample_weight)

            # store each fitted tree in a dictionary
            self.tree_dic[col] = tree

            # create the bins series
            # predict, group by original values
            # create monotonicly increasing bin for pd.cut
            X[f"{col}_g"] = tree.predict(X[[col]])

            if (self.cat_features is not None) and (col in self.cat_features):
                self.cat_bin_dict[col] = (
                    X[[f"{col}_g", col]]
                    .groupby(f"{col}_g")
                    .aggregate(
                        lambda x: " / ".join(
                            map(
                                str,
                                encoder.inverse_transform(
                                    np.expand_dims(X.loc[x.index, col].unique(), axis=1)
                                ).ravel(),
                            )
                        )
                    )
                    .to_dict()[col]
                )

            else:
                bin_array = (
                    X[[f"{col}_g", col]]
                    .groupby(f"{col}_g")
                    .aggregate(max)
                    .sort_values(col)
                    .values.ravel()
                )
                # overwrite max for handling unseen larger values than max val on the train set
                bin_array = np.delete(bin_array, [np.argmax(bin_array)])
                # append -Inf and Inf for covering the whole interval, as in KBinsDiscretizer
                bin_array = np.unique(np.append(bin_array, [-np.Inf, np.Inf]))
                self.bin_upper_bound_dic[col] = bin_array
                # the tree imputer: if nan is passed, the predicted value will be the
                # same than another non-missing value and nan should fall into this bin
                # the predicted value of a NaN
                nan_pred_val = tree.predict(np.expand_dims([np.nan], axis=1))[0]
                # the closest predicted value for non NaN inputs
                non_nan_values = X[col].dropna().unique()
                pred_values = tree.predict(np.expand_dims(non_nan_values, axis=1))
                # store the value for knowing the bin into which the NaN should fall
                self.tree_imputer[col] = non_nan_values.flat[
                    np.abs(pred_values - nan_pred_val).argmin()
                ]

            del tree
        return self

    def transform(self, X):
        """Apply the discretizer on `X`.
        Parameters
        ----------
        X :
            Input data shape (n_samples, n_features), where `n_samples` is the number of samples and
            `n_features` is the number of features.


        Returns
        -------
        X :
            pd.DataFrame with the binned and grouped columns
        """
        X = X.copy()
        for col in self.bin_features:
            if self.raw:
                # predict each univariate tree
                X[col] = self.tree_dic[col].predict(X[[col]])
            else:
                if (self.cat_features is not None) and (col in self.cat_features):
                    # apply the systematic imputation (missing might be grouped
                    # with other categories depending on the results of the tree
                    # splitting)
                    X[col] = (
                        X[col]
                        .astype("category")
                        .cat.add_categories("missing_added")
                        .fillna("missing_added")
                    )
                    dum = self.ordinal_encoder_dic[col].transform(X[[col]])
                    X[col] = self.tree_dic[col].predict(dum)
                    X[col] = X[col].map(self.cat_bin_dict[col])
                else:
                    # retrieve the association the tree learnt for missing values
                    X[col].fillna(self.tree_imputer[col], inplace=True)
                    # apply the binning
                    X[col] = pd.cut(
                        X[col],
                        bins=self.bin_upper_bound_dic[col],
                        include_lowest=True,
                        precision=1,
                    )
        return X


def highlight_discarded(s):
    """highlight X in red and V in green.

    Parameters
    ----------
    s : np.arrays

    Returns
    -------
    list

    """
    is_X = s == 0
    return [
        "background-color: #d65f5f" if v else "background-color: #33a654" for v in is_X
    ]


def make_fs_summary(selector_pipe):
    """make_fs_summary makes a summary dataframe highlighting at which step a
    given predictor has been rejected (if any).

    Parameters
    ----------
    selector_pipe : sklearn.pipeline.Pipeline
        the feature selector pipeline.

    Examples
    --------
    >>> groot_pipeline = Pipeline([
    ... ('missing', MissingValueThreshold()),
    ... ('unique', UniqueValuesThreshold()),
    ... ('cardinality', CardinalityThreshold()),
    ... ('collinearity', CollinearityThreshold(threshold=0.5)),
    ... ('lowimp', VariableImportance(eval_metric='poisson', objective='poisson', verbose=2)),
    ... ('grootcv', GrootCV(objective='poisson', cutoff=1, n_folds=3, n_iter=5))])
    >>> groot_pipeline.fit_transform(
        X=df[predictors],
        y=df[target],
        lowimp__sample_weight=df[weight],
        grootcv__sample_weight=df[weight])
    >>> fs_summary_df = make_fs_summary(groot_pipeline)
    """

    tag_df = pd.DataFrame({"predictor": selector_pipe[0].feature_names_in_})
    for selector_name in selector_pipe.named_steps.keys():
        if hasattr(selector_pipe.named_steps[selector_name], "feature_names_in_"):
            feature_in = selector_pipe.named_steps[selector_name].feature_names_in_
            to_drop = list(
                set(selector_pipe.named_steps[selector_name].feature_names_in_)
                - set(selector_pipe.named_steps[selector_name].get_feature_names_out())
            )
            tag_df[selector_name] = np.where(
                tag_df["predictor"].isin(to_drop), 0, 1
            ) * np.where(tag_df["predictor"].isin(feature_in), 1, np.nan)

    col_to_apply_style = tag_df.columns[1:]
    tag_df = (
        tag_df.style.apply(highlight_discarded, subset=col_to_apply_style)
        .applymap(lambda x: "" if x == x else "background-color: #ffa500")
        .format(precision=0)
    )
    return tag_df



"""This module provide methods for sampling large datasets for reducing the running time
    
"""
import numpy as np
import pandas as pd
from scipy.sparse import issparse
from collections import Counter
from sklearn.cluster import AgglomerativeClustering
from sklearn.ensemble import IsolationForest
from scipy.stats import ks_2samp
from .utils import is_list_of_str, is_list_of_bool, is_list_of_int


def sample(df, n=1000, sample_weight=None, method="gower"):
    """Sampling rows from a dataframe when random sampling is not
    enough for reducing the number of rows.
    The strategies are either using hierarchical clustering
    based on the Gower distance or using isolation forest for identifying
    the most similar elements.
    For the clustering algorithm, clusters are determined using the Gower distance
    (mixed type data) and the dataset is shrunk from n_samples to n_clusters.

    For the isolation forest algorithm, samples are added till a suffisant 2-samples
    KS statistics is reached or if the number iteration reached the max number (20)

    Parameters
    ----------
    df : pd.DataFrame
        the dataframe to sample, with or without the target
    n : int, optional
        the number of clusters if method is ``"gower"``, by default 100
    sample_weight : pd.Series or np.array, optional
        sample weights, by default None
    method : str, optional
        the strategy to use for sampling the rows. Either ``"gower"`` or ``"isoforest"``, by default 'gower'

    Returns
    -------
    pd.DataFrame
        the sampled dataframe

    """
    assert isinstance(df, pd.DataFrame), "X should be a DataFrame"
    X = df.copy()
    num_cols = list(X.select_dtypes(include=[np.number]))
    non_num_cols = list(set(list(X.columns)) - set(num_cols))

    if method == "gower":
        # basic imputation
        if non_num_cols:
            X[non_num_cols] = X[non_num_cols].fillna(X[non_num_cols].mode().iloc[0])
        if num_cols:
            X[num_cols] = X[num_cols].fillna(X[num_cols].mean().iloc[0])

        # no need for scaling, it is built-in the computation of the Gower distance
        gd = gower_matrix(X, cat_features=non_num_cols, weight=sample_weight)

        labels = AgglomerativeClustering(
            n_clusters=n, affinity="precomputed", linkage="complete"
        ).fit_predict(gd)
        X["label"] = labels
        X["label"] = "clus_" + X["label"].astype(str)
        X_num = X.groupby("label")[num_cols].agg("mean")
        if non_num_cols:
            X_nonnum = X.groupby("label")[non_num_cols].agg(get_most_common)
            X_sampled = X_num.join(X_nonnum)
        else:
            X_sampled = X_num
        X_sampled = X_sampled.reindex(X.columns, axis=1)
        return X_sampled
    elif method == "isoforest":
        X[non_num_cols] = X[non_num_cols].astype("str").astype("category")
        for col in non_num_cols:
            X[col] = X[col].astype("category").cat.codes
        idx = isof_find_sample(X, sample_weight=None)
        return X.iloc[idx, :]
    else:
        NotImplementedError(f"{method} not implemented")


def get_most_common(srs):
    x = list(srs)
    my_counter = Counter(x)
    return my_counter.most_common(1)[0][0]


def gower_matrix(
    data_x,
    data_y=None,
    weight=None,
    cat_features="auto",
):
    """Computes the gower distances between X and Y

    Gower is a similarity measure for categorical, boolean and numerical mixed
    data.

    Parameters
    ----------
    data_x : np.array or pd.DataFrame
        The data for computing the Gower distance
    data_y : np.array or pd.DataFrame or pd.Series, optional
        The reference matrix or vector to compare with, optional
    weight : np.array or pd.Series, optional
        sample weight, optional
    cat_features : list of str or bool or int, optional
        auto-detect cat features or a list of cat features, by default 'auto'

    Returns
    -------
    np.array
        The Gower distance matrix, shape (n_samples, n_samples)

    Notes
    -----
    The non-numeric features, and numeric feature ranges are determined from X and not Y.

    Raises
    ------
    TypeError
        If two dataframes are passed but have different number of columns
    TypeError
        If two arrays are passed but have different number of columns
    TypeError
        Sparse matrices are not supported
    TypeError
        if a list of categorical columns is passed, it should be a list of strings or integers or boolean values
    """
    # function checks
    X = data_x
    if data_y is None:
        Y = data_x
    else:
        Y = data_y
    if not isinstance(X, np.ndarray):
        y_col = Y.columns if isinstance(Y, pd.DataFrame) else Y.index
        if not np.array_equal(X.columns, y_col):
            raise TypeError("X and Y must have same columns!")
    else:
        if not X.shape[1] == Y.shape[1]:
            raise TypeError("X and Y must have same y-dim!")
    if issparse(X) or issparse(Y):
        raise TypeError("Sparse matrices are not supported!")

    x_n_rows, x_n_cols = X.shape
    y_n_rows, y_n_cols = Y.shape

    if cat_features == "auto":
        if not isinstance(X, np.ndarray):
            is_number = np.vectorize(lambda x: not np.issubdtype(x, np.number))
            cat_features = is_number(X.dtypes)
        else:
            cat_features = np.zeros(x_n_cols, dtype=bool)
            for col in range(x_n_cols):
                if not np.issubdtype(type(X[0, col]), np.number):
                    cat_features[col] = True
    else:
        # force categorical columns (if integer encoded for instance)
        if is_list_of_str(cat_features):
            cat_feat = [True if c in cat_features else False for c in X.columns]
            cat_features = np.array(cat_feat)
        elif is_list_of_bool(cat_features):
            cat_features = np.array(cat_features)
        elif is_list_of_int(cat_features):
            cat_feat = [
                True if c in cat_features else False for c in range(len(X.columns))
            ]
            cat_features = np.array(cat_feat)
        else:
            raise TypeError(
                "If not 'auto' cat_features should be a list of strings, integers or Booleans"
            )

    # print(cat_features)

    if not isinstance(X, np.ndarray):
        X = np.asarray(X)
    if not isinstance(Y, np.ndarray):
        Y = np.asarray(Y)

    Z = np.concatenate((X, Y))

    x_index = range(0, x_n_rows)
    y_index = range(x_n_rows, x_n_rows + y_n_rows)

    Z_num = Z[:, np.logical_not(cat_features)]

    num_cols = Z_num.shape[1]
    num_ranges = np.zeros(num_cols)
    num_max = np.zeros(num_cols)

    for col in range(num_cols):
        col_array = Z_num[:, col].astype(np.float32)
        max_ = np.nanmax(col_array)
        min_ = np.nanmin(col_array)

        if np.isnan(max_):
            max_ = 0.0
        if np.isnan(min_):
            min_ = 0.0
        num_max[col] = max_
        num_ranges[col] = (1 - min_ / max_) if (max_ != 0) else 0.0

    # This is to normalize the numeric values between 0 and 1.
    Z_num = np.divide(
        Z_num.astype(float),
        num_max.astype(float),
        out=np.zeros_like(Z_num).astype(float),
        where=num_max != 0,
    )
    Z_cat = Z[:, cat_features]

    if weight is None:
        weight = np.ones(Z.shape[1])

    # print(weight)

    weight_cat = weight[cat_features]
    weight_num = weight[np.logical_not(cat_features)]

    out = np.zeros((x_n_rows, y_n_rows), dtype=np.float32)

    weight_sum = weight.sum()

    X_cat = Z_cat[
        x_index,
    ]
    X_num = Z_num[
        x_index,
    ]
    Y_cat = Z_cat[
        y_index,
    ]
    Y_num = Z_num[
        y_index,
    ]

    # print(X_cat,X_num,Y_cat,Y_num)

    for i in range(x_n_rows):
        j_start = i
        if x_n_rows != y_n_rows:
            j_start = 0
        # call the main function
        res = _gower_distance_row(
            X_cat[i, :],
            X_num[i, :],
            Y_cat[j_start:y_n_rows, :],
            Y_num[j_start:y_n_rows, :],
            weight_cat,
            weight_num,
            weight_sum,
            num_ranges,
        )
        # print(res)
        out[i, j_start:] = res
        if x_n_rows == y_n_rows:
            out[i:, j_start] = res

    return out


def _gower_distance_row(
    xi_cat,
    xi_num,
    xj_cat,
    xj_num,
    feature_weight_cat,
    feature_weight_num,
    feature_weight_sum,
    ranges_of_numeric,
):
    """Compute a row of the Gower matrix

    Parameters
    ----------
    xi_cat : np.array
        categorical row of the X matrix
    xi_num : np.array
        numerical row of the X matrix
    xj_cat : np.array
        categorical row of the X matrix
    xj_num : np.array
        numerical row of the X matrix
    feature_weight_cat : np.array
        weight vector for the categorical features
    feature_weight_num : np.array
        weight vector for the numerical features
    feature_weight_sum : float
        The sum of the wieghts
    ranges_of_numeric : np.array
        range of the scaled numerical features (between 0 and 1)

    Returns
    -------
    np.array : array
        a row vector of the Gower distance
    """
    # categorical columns
    sij_cat = np.where(xi_cat == xj_cat, np.zeros_like(xi_cat), np.ones_like(xi_cat))
    sum_cat = np.multiply(feature_weight_cat, sij_cat).sum(axis=1)

    # numerical columns
    abs_delta = np.absolute(xi_num - xj_num)
    sij_num = np.divide(
        abs_delta,
        ranges_of_numeric,
        out=np.zeros_like(abs_delta),
        where=ranges_of_numeric != 0,
    )

    sum_num = np.multiply(feature_weight_num, sij_num).sum(axis=1)
    sums = np.add(sum_cat, sum_num)
    sum_sij = np.divide(sums, feature_weight_sum)

    return sum_sij


def smallest_indices(ary, n):
    """Returns the n largest indices from a numpy array.

    Parameters
    ----------
    ary : np.array
        the array for which to return largest indices
    n : int
        the number of indices to return

    Returns
    -------
    dict
        the dictionary of indices and values of the largest elements
    """
    # n += 1
    flat = np.nan_to_num(ary.flatten(), nan=999)
    indices = np.argpartition(-flat, -n)[-n:]
    indices = indices[np.argsort(flat[indices])]
    # indices = np.delete(indices,0,0)
    values = flat[indices]
    return {"index": indices, "values": values}


def gower_topn(
    data_x,
    data_y=None,
    weight=None,
    cat_features="auto",
    n=5,
    key=None,
):
    """Get the n most similar elements

    Parameters
    ----------
    data_x : np.array or pd.DataFrame
        The data for the look up
    data_y : np.array or pd.DataFrame or pd.Series, optional
        elements for which to return the most similar elements, should be a single row
    weight : np.array or pd.Series, optional
        sample weight, by default None
    cat_features : list of str or bool or int, optional
        auto detection of cat features or a list of strings, booleans or integers, by default 'auto'
    n : int, optional
        the number of neighbors/similar rows to find, by default 5
    key : str, optional
        identifier key. If several rows refer to the same id, this column
        will be used for finding the nearest neighbors with a
        different id, by default None

    Returns
    -------
    dict
        the dictionary of indices and values of the closest elements

    Raises
    ------
    TypeError
        if the reference element is not a single row
    """

    if data_y.shape[0] >= 2:
        raise TypeError("Only support `data_y` of 1 row. ")
    if key is None:
        dm = gower_matrix(data_y, data_x, weight, cat_features)
    else:
        X = data_x.drop(key, axis=1)
        Y = data_x.drop(key, axis=1)
        dm = gower_matrix(Y, X, weight, cat_features)

    if key is not None:
        idx = smallest_indices(np.nan_to_num(dm[0], nan=1), n)["index"]
        val = smallest_indices(np.nan_to_num(dm[0], nan=1), n)["values"]
        unique_id = data_x.iloc[idx, :]
        unique_id = unique_id[key]
        nunique_id = unique_id.nunique()
        mul = 1
        # continue looking for the closest n unique records with a different id
        while nunique_id < n:
            idx = smallest_indices(np.nan_to_num(dm[0], nan=1), mul * n)["index"]
            val = smallest_indices(np.nan_to_num(dm[0], nan=1), mul * n)["values"]
            unique_id = data_x.iloc[idx, :].reset_index()
            unique_id = unique_id[key]
            nunique_id = unique_id.nunique()
            mul += 1

        # find the indices of the unique id
        _, idx_n = np.unique(unique_id, return_index=True)
        # select only the rows corresponding to unique id
        val = val[idx_n]
        idx = idx[idx_n]
        # sort them from the closest to the farthest, according to the Gower metrics
        idx_n = np.argsort(val)
        # return the n closest records, with a different id
        return {"index": idx[idx_n[:n]], "values": val[idx_n[:n]]}
    else:
        return smallest_indices(np.nan_to_num(dm[0], nan=1), n)


def get_5_percent_splits(length):
    """splits dataframe into 5% intervals

    Parameters
    ----------
    length : int
        array length

    Returns
    -------
    array
        vector of sizes
    """

    five_percent = round(5 / 100 * length)
    return np.arange(five_percent, length, five_percent)


def isolation_forest(X, sample_weight=None):
    """fits isloation forest to the dataset and gives an anomally score to every sample

    Parameters
    ----------
    X : pd.DataFrame or np.array
        the predictors matrix
    sample_weight : pd.Series or np.array, optional
        the sample weights, if any, by default None
    """
    clf = IsolationForest().fit(X, sample_weight=sample_weight)
    return clf.score_samples(X)


def isof_find_sample(X, sample_weight=None):
    """Finds a sample by comparing the distributions of the anomally scores between the sample and the original
    distribution using the KS-test. Starts of a 5% howver will increase to 10% and then 15% etc. if a significant sample can not be found

    References
    ----------
    Sampling method taken from boruta_shap, author: https://github.com/Ekeany

    Parameters
    ----------
    X : pd.DataFrame
        the predictors matrix
    sample_weight : pd.Series or np.array, optional
        the sample weights, if any, by default None

    Returns
    -------
    array
        the indices for reducing the shadow predictors matrix
    """
    loop = True
    iteration = 0
    size = get_5_percent_splits(length=X.shape[0])
    element = 1
    preds = isolation_forest(X, sample_weight)
    while loop:
        sample_indices = np.random.choice(
            np.arange(preds.size), size=size[element], replace=False
        )
        sample = np.take(preds, sample_indices)
        if ks_2samp(preds, sample).pvalue > 0.95:
            break
        if iteration == 20:
            element += 1
            iteration = 0
    return sample_indices



"""
This module provides utilities for parallelizing operations on pd.DataFrame

**The module structure is the following:**

- The ``parallel_matrix_entries`` for parallelizing operations returning a matrix (2D) (apply on pairs of columns)
- The ``parallel_df`` for parallelizing operations returning a series (1D) (apply on a single column at a time)
"""

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from multiprocessing import cpu_count
from itertools import chain


def parallel_matrix_entries(func, df, comb_list, sample_weight=None, n_jobs=-1):
    """parallel_matrix_entries applies a function to each chunk of
    combinaison of columns of the dataframe, distributed by cores.
    This is similar to https://github.com/smazzanti/mrmr/mrmr/pandas.py


    Parameters
    ----------
    func : callable
        function to be applied to each column
    df : pd.DataFrame
        the dataframe on which to apply the function
    comb_list : list of tuples of str
        Pairs of column names corresponding to the entries
    sample_weight : pd.Series or np.array, optional
        The weight vector, if any, of shape (n_samples,), by default None
    n_jobs : int, optional
        the number of cores to use for the computation, by default -1

    Returns
    -------
    pd.DataFrame
        concatenated results into a single pandas DF
    """

    if n_jobs == 1:
        return func(X=df, sample_weight=sample_weight, comb_list=comb_list)

    n_jobs = (
        min(cpu_count(), len(df.columns)) if n_jobs == -1 else min(cpu_count(), n_jobs)
    )
    comb_chunks = np.array_split(comb_list, n_jobs)
    lst = Parallel(n_jobs=n_jobs)(
        delayed(func)(X=df, sample_weight=sample_weight, comb_list=comb_chunk)
        for comb_chunk in comb_chunks
    )
    # return flatten list of pandas DF
    return pd.concat(list(chain(*lst)), ignore_index=True)


def parallel_df(func, df, series, sample_weight=None, n_jobs=-1):
    """parallel_df apply a function to each column of the dataframe, distributed by cores.
    This is similar to https://github.com/smazzanti/mrmr/mrmr/pandas.py

    Parameters
    ----------
    func : callable
        function to be applied to each column
    df : pd.DataFrame
        the dataframe on which to apply the function
    series : pd.Series
        series (target) used by the function
    sample_weight : pd.Series or np.array, optional
        The weight vector, if any, of shape (n_samples,), by default None
    n_jobs : int, optional
        the number of cores to use for the computation, by default -1

    Returns
    -------
    pd.DataFrame
        concatenated results into a single pandas DF
    """

    if n_jobs == 1:
        return func(df, series, sample_weight).sort_values(ascending=False)

    n_jobs = (
        min(cpu_count(), len(df.columns)) if n_jobs == -1 else min(cpu_count(), n_jobs)
    )
    col_chunks = np.array_split(range(len(df.columns)), n_jobs)
    lst = Parallel(n_jobs=n_jobs)(
        delayed(func)(df.iloc[:, col_chunk], series, sample_weight)
        for col_chunk in col_chunks
    )
    return pd.concat(lst).sort_values(ascending=False)


def _compute_series(
    X,
    y,
    sample_weight=None,
    func_xyw=None,
):
    """_compute_series is a utility function for computing the series
    resulting of the ``apply``

    Parameters
    ----------
    X : pd.DataFrame, of shape (n_samples, n_features)
        The set of regressors that will be tested sequentially
    y : pd.Series or np.array, of shape (n_samples,)
        The target vector
    sample_weight : pd.Series or np.array, of shape (n_samples,), optional
        The weight vector, if any, by default None
    func_xyw : callable, optional
        callable (function) for computing the individual elements of the series
        takes two mandatory inputs (x and y) and an optional input w, sample_weights
    """

    def _closure_compute_series(x, y, sample_weight):
        x_not_na = ~x.isna()
        if x_not_na.sum() == 0:
            return 0
        return func_xyw(
            x=x[x_not_na],
            y=y[x_not_na],
            sample_weight=sample_weight[x_not_na],
            as_frame=False,
        )

    return X.apply(
        lambda col: _closure_compute_series(x=col, y=y, sample_weight=sample_weight)
    ).fillna(0.0)


def _compute_matrix_entries(
    X,
    comb_list,
    sample_weight=None,
    func_xyw=None,
):
    """base closure for computing matrix entries appling a function to each chunk of
    combinaison of columns of the dataframe, distributed by cores.
    This is similar to https://github.com/smazzanti/mrmr/mrmr/pandas.py

    Parameters
    ----------
    X : pd.DataFrame, of shape (n_samples, n_features)
        The set of regressors that will be tested sequentially
    sample_weight : pd.Series or np.array, of shape (n_samples,), optional
        The weight vector, if any, by default None
    func_xyw : callable, optional
        callable (function) for computing the individual elements of the matrix
        takes two mandatory inputs (x and y) and an optional input w, sample_weights
    comb_list : list of 2-uple of str
        Pairs of column names corresponding to the entries

    Returns
    -------
    pd.DataFrame
        concatenated results into a single pandas DF
    """
    v_df_list = []
    for comb in comb_list:
        v_df_list.append(
            func_xyw(
                x=X[comb[0]],
                y=X[comb[1]],
                sample_weight=sample_weight,
                as_frame=True,
            )
        )

    return v_df_list



"""
Utility and validation functions
  
"""

from __future__ import print_function, division

import lightgbm as lgb
import matplotlib as mpl
import numpy as np
import pandas as pd

from pkg_resources import resource_filename
from matplotlib import pyplot as plt
from sklearn.impute import SimpleImputer
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import Pipeline
from sklearn.datasets import fetch_openml
from sklearn.datasets import load_breast_cancer
from sklearn.utils import Bunch
import joblib

qualitative_colors = [
    "#7F3C8D",
    "#11A579",
    "#3969AC",
    "#F2B701",
    "#E73F74",
    "#80BA5A",
    "#E68310",
    "#008695",
    "#CF1C90",
    "#F97B72",
]

#####################
#                   #
#     Utilities     #
#                   #
#####################


def reset_plot():
    """Reset plot style"""
    # plt.rcParams = plt.rcParamsDefault
    mpl.rcParams.update(plt.rcParamsDefault)


def set_my_plt_style(height=3, width=5, linewidth=2):
    """This set the style of matplotlib to fivethirtyeight with some modifications (colours, axes)

    Parameters
    ----------
    linewidth: int, default=2
        line width
    height: int, default=3
        fig height in inches (yeah they're still struggling with the metric system)
    width: int, default=5
        fig width in inches (yeah they're still struggling with the metric system)

    """
    plt.style.use("fivethirtyeight")
    my_colors_list = qualitative_colors
    myorder = [2, 3, 4, 1, 0, 6, 5, 8, 9, 7]
    my_colors_list = [my_colors_list[i] for i in myorder]
    bckgnd_color = "#f5f5f5"
    params = {
        "figure.figsize": (width, height),
        "axes.prop_cycle": plt.cycler(color=my_colors_list),
        "axes.facecolor": bckgnd_color,
        "patch.edgecolor": bckgnd_color,
        "figure.facecolor": bckgnd_color,
        "axes.edgecolor": bckgnd_color,
        "savefig.edgecolor": bckgnd_color,
        "savefig.facecolor": bckgnd_color,
        "grid.color": "#9e9e9e",
        "lines.linewidth": linewidth,
    }  # plt.cycler(color=my_colors_list)
    mpl.rcParams.update(params)


def create_dtype_dict(df: pd.DataFrame, dic_keys: str = "col_names"):
    """create a custom dictionary of data type for adding suffixes
    to column names in the plotting utility for association matrix

    Parameters
    ----------
    df :
        the dataframe used for computing the association matrix
    dic_keys :
        Either "col_names" or "dtypes" for returning either a dictionary
        with column names or dtypes as keys.
    """
    cat_cols = list(df.select_dtypes(include=["object", "category", "bool"]))
    time_cols = list(df.select_dtypes(include=["datetime", "timedelta", "datetimetz"]))
    num_cols = list(df.select_dtypes(include=[np.number]))
    remaining_cols = list(
        set(df.columns) - set(cat_cols).union(set(num_cols)).union(time_cols)
    )

    if dic_keys == "col_names":
        cat_dic = {c: "cat" for c in cat_cols}
        num_dic = {c: "num" for c in num_cols}
        time_dic = {c: "time" for c in time_cols}
        remainder_dic = {c: "unk" for c in remaining_cols}
        return {**cat_dic, **num_dic, **time_dic, **remainder_dic}
    elif dic_keys == "dtypes":
        cat_dic = {"cat": cat_cols}
        num_dic = {"num": num_cols}
        time_dic = {"time": time_cols}
        remainder_dic = {"unk": remaining_cols}
        return {**cat_dic, **num_dic, **time_dic, **remainder_dic}
    else:
        raise ValueError("'dic_keys' should be either 'col_names' or 'dtypes'")


def get_pandas_cat_codes(X):
    dtypes_dic = create_dtype_dict(X, dic_keys="dtypes")
    obj_feat = dtypes_dic["cat"] + dtypes_dic["time"] + dtypes_dic["unk"]

    if obj_feat:
        cat = X[obj_feat].stack().astype("str").astype("category").cat.codes.unstack()
        X = pd.concat([X[X.columns.difference(obj_feat)], cat], axis=1)
        cat_idx = [X.columns.get_loc(col) for col in obj_feat]
    else:
        obj_feat = None
        cat_idx = None

    return X, obj_feat, cat_idx


def check_if_tree_based(model):
    """check if estimator is tree based

    Parameters
    ----------
    model : object
        the estimator to check

    Returns
    -------
    condition : boolean
        if tree based or not
    """
    tree_based_models = ["lightgbm", "xgboost", "catboost", "_forest", "boosting"]
    condition = any(i in str(type(model)).lower() for i in tree_based_models)
    return condition


def is_lightgbm(estimator):
    """check if estimator is lightgbm

    Parameters
    ----------
    model : object
        the estimator to check

    Returns
    -------
    condition : boolean
        if lgbm based or not
    """
    is_lgb = "lightgbm" in str(type(estimator))
    return is_lgb


def is_catboost(estimator):
    """check if estimator is catboost

    Parameters
    ----------
    model : object
        the estimator to check

    Returns
    -------
    condition : boolean
        if catboost based or not
    """
    is_cat = "catboost" in str(type(estimator))
    return is_cat


def is_xgboost(estimator):
    """check if estimator is xgboost

    Parameters
    ----------
    model : object
        the estimator to check

    Returns
    -------
    condition : boolean
        if xgboost based or not
    """
    is_xgb = "xgboost" in str(type(estimator))
    return is_xgb


def LightForestRegressor(n_feat, n_estimators=10):
    """lightGBM implementation of the Random Forest regressor with the
    ideal number of features, according to Elements of statistical learning

    Parameters
    ----------
    n_feat: int
        the number of predictors (nbr of columns of the X matrix)
    n_estimators : int, optional
        the number of trees/estimators, by default 10

    Returns
    -------
    lightgbm regressor
        sklearn random forest estimator based on lightgbm
    """

    feat_frac = n_feat / (3 * n_feat)
    return lgb.LGBMRegressor(
        verbose=-1,
        force_col_wise=True,
        n_estimators=n_estimators,
        subsample=0.632,
        colsample_bytree=feat_frac,
        boosting_type="rf",
        subsample_freq=1,
    )


def LightForestClassifier(n_feat, n_estimators=10):
    """lightGBM implementation of the Random Forest classifier with the
    ideal number of features, according to Elements of statistical learning

    Parameters
    ----------
    n_feat: int
        the number of predictors (nbr of columns of the X matrix)
    n_estimators : int, optional
        the number of trees/estimators, by default 10

    Returns
    -------
    lightgbm classifier
        sklearn random forest estimator based on lightgbm
    """
    feat_frac = np.sqrt(n_feat) / n_feat
    return lgb.LGBMClassifier(
        verbose=-1,
        force_col_wise=True,
        n_estimators=n_estimators,
        subsample=0.632,
        colsample_bytree=feat_frac,
        boosting_type="rf",
        subsample_freq=1,
    )


def is_list_of_str(str_list):
    """Check if ``str_list`` is not a list of strings

    Parameters
    ----------
    str_list : list of str
        the list we want to check for

    Returns
    -------
    bool
        True if list of strings, else False
    """
    if str_list is not None:
        if not (
            isinstance(str_list, list) and all(isinstance(s, str) for s in str_list)
        ):
            return False
        else:
            return True


def is_list_of_bool(bool_list):
    """Check if ``bool_list`` is not a list of Booleans

    Parameters
    ----------
    bool_list : list of bool
        the list we want to check for

    Returns
    -------
    bool
        True if list of Booleans, else False
    """
    if bool_list is not None:
        if not (
            isinstance(bool_list, list) and all(isinstance(s, bool) for s in bool_list)
        ):
            return False
        else:
            return True


def is_list_of_int(int_list):
    """Check if ``int_list`` is not a list of integers

    Parameters
    ----------
    int_list : list of int
        the list we want to check for

    Returns
    -------
    bool
        True if list of integers, else False
    """
    if int_list is not None:
        if not (
            isinstance(int_list, list) and all(isinstance(s, int) for s in int_list)
        ):
            return False
        else:
            return True


def _get_titanic_data():
    """Load Titanic data and add dummies (random predictors, numeric and categorical) and
    a genuine one, for benchmarking purpose. Classification (binary)

    Returns
    -------
    object
        Bunch sklearn, extension of dictionary
    """
    # Fetch Titanic data and add random cat and numbers
    # Example taken from https://scikit-learn.org/stable/auto_examples/inspection/
    # plot_permutation_importance.html#sphx-glr-auto-examples-inspection-plot-permutation-importance-py
    X, y = fetch_openml("titanic", version=1, as_frame=True, return_X_y=True)
    rng = np.random.RandomState(seed=42)
    nice_guys = ["Rick", "Bender", "Cartman", "Morty", "Fry", "Vador", "Thanos"]
    X["random_cat"] = np.random.choice(nice_guys, X.shape[0])
    X["random_num"] = rng.randn(X.shape[0])
    X["family_size"] = X["parch"] + X["sibsp"]
    X.drop(["parch", "sibsp"], axis=1, inplace=True)
    X["is_alone"] = 1
    X.loc[X["family_size"] > 1, "is_alone"] = 0
    X["title"] = (
        X["name"].str.split(", ", expand=True)[1].str.split(".", expand=True)[0]
    )
    X.loc[X.title == "Miss", "title"] = "Mrs"
    title_counts = X["title"].value_counts()
    rare_titles = title_counts[title_counts < 10].index
    X.loc[X["title"].isin(rare_titles), "title"] = "rare"
    # X['title'] = X.title.apply(lambda x: 'rare' if rare_titles[x] else x)

    categorical_columns = [
        "pclass",
        "sex",
        "embarked",
        "random_cat",
        "is_alone",
        "title",
    ]
    numerical_columns = ["age", "family_size", "fare", "random_num"]
    X = X[categorical_columns + numerical_columns]
    # Impute
    categorical_pipe = Pipeline(
        [("imputer", SimpleImputer(strategy="constant", fill_value="missing"))]
    )
    numerical_pipe = Pipeline([("imputer", SimpleImputer(strategy="mean"))])
    preprocessing = ColumnTransformer(
        [
            ("cat", categorical_pipe, categorical_columns),
            ("num", numerical_pipe, numerical_columns),
        ]
    )
    X_trans = preprocessing.fit_transform(X)
    X = pd.DataFrame(X_trans, columns=X.columns)
    # encode
    # X, cat_var_df, inv_mapper = cat_var(X)
    X[categorical_columns] = X[categorical_columns].astype(str)
    X[numerical_columns] = X[numerical_columns].astype(float)
    # sample weight is just a dummy random vector for testing purpose
    sample_weight = np.random.uniform(0, 1, len(y))
    return Bunch(
        data=X, target=y, sample_weight=sample_weight, categorical=categorical_columns
    )


def _get_cancer_data():
    """Load breast cancer data and add dummies (random predictors) and a genuine one, for benchmarking purpose
    Classification (binary)

    Returns
    -------
    object
        Bunch sklearn, extension of dictionary
    """

    rng = np.random.RandomState(seed=42)
    data = load_breast_cancer()
    X, y = data.data, data.target
    X = pd.DataFrame(X)
    X.columns = data.feature_names
    X["random_num1"] = rng.randn(X.shape[0])
    X["random_num2"] = np.random.poisson(1, X.shape[0])
    z = y.astype(int)
    X["genuine_num"] = z * np.abs(
        np.random.normal(0, 0.1, X.shape[0])
    ) + np.random.normal(0, 0.1, X.shape[0])
    y = pd.Series(y)
    return Bunch(data=X, target=y, sample_weight=None, categorical=None)


def _load_boston_data():
    """Load Boston data and add dummies (random predictors, numeric and categorical) and
    a genuine one, for benchmarking purpose. Regression (positive domain).

    Returns
    -------
    object
        Bunch sklearn, extension of dictionary
    """

    data_file_name = resource_filename(__name__, "dataset/data/boston_bunch.joblib")
    return joblib.load(data_file_name)


def _load_housing(as_frame: bool = False):
    """Load the California housing data. See here
    https://scikit-learn.org/stable/modules/generated/sklearn.datasets.fetch_california_housing.html
    for the downloadable version.

    Parameters
    ----------
    as_frame :
        return a pandas dataframe? if not then a "Bunch" (enhanced dictionary) is returned (default ``True``)

    Returns
    -------
    pd.DataFrame or Bunch
        the dataset

    """
    fdescr_name = resource_filename(__name__, "dataset/descr/housing.rst")
    with open(fdescr_name) as f:
        descr_text = f.read()

    data_file_name = resource_filename(__name__, "dataset/data/housing.zip")
    data = pd.read_csv(data_file_name)
    feature_names = [
        "MedInc",
        "HouseAge",
        "AveRooms",
        "AveBedrms",
        "Population",
        "AveOccup",
        "Latitude",
        "Longitude",
    ]

    if as_frame:
        return data
    else:
        return Bunch(
            data=data[feature_names].values,
            target=data["target"].values,
            feature_names=feature_names,
            DESCR=descr_text,
            filename=data_file_name,
        )


def plot_y_vs_X(X, y, ncols=2, figsize=(10, 10)):
    """Plot target vs relevant and non-relevant predictors

    Parameters
    ----------
    X : pd.DataFrame
        the pd DF of the predictors
    y : np.array
        the target
    ncols : int, optional
        the number of columns in the facet plot, by default 2
    figsize : tuple, optional
        the figure size, by default (10, 10)

    Returns
    -------
    plt.figure
        the univariate plots y vs pred_i
    """

    X = pd.DataFrame(X)
    ncols_to_plot = X.shape[1]
    n_rows = int(np.ceil(ncols_to_plot / ncols))

    # Create figure and axes (this time it's 9, arranged 3 by 3)
    f, axs = plt.subplots(nrows=n_rows, ncols=ncols, figsize=figsize)

    # delete non-used axes
    n_charts = ncols_to_plot
    n_subplots = n_rows * ncols
    cols_to_enum = X.columns

    # Make the axes accessible with single indexing
    if n_charts > 1:
        axs = axs.flatten()

    for i, col in enumerate(cols_to_enum):
        # select the axis where the map will go
        if n_charts > 1:
            ax = axs[i]
        else:
            ax = axs

        ax.scatter(X[col], y, alpha=0.1)
        ax.set_title(col)

    if n_subplots > n_charts > 1:
        for i in range(n_charts, n_subplots):
            ax = axs[i]
            ax.set_axis_off()

    # Display the figure
    plt.tight_layout()
    return f


def load_data(name="Titanic"):
    """Load some toy data set to test the All Relevant Feature Selection methods.
    Dummies (random) predictors are added and ARFS should be able to filter them out.
    The Titanic predictors are encoded (needed for scikit estimators).

    Titanic and cancer are for binary classification, they contain synthetic random (dummies) predictors and a
    noisy but genuine synthetic predictor. Hopefully, a good All Relevant FS should be able to detect all the
    predictors genuinely related to the target.

    Boston is for regression, this data set contains

    Parameters
    ----------
    name : str, optional
        the name of the data set. Titanic is for classification with sample_weight,
        Boston for regression and cancer for classification (without sample weight), by default 'Titanic'

    Returns
    -------
    Bunch
        extension of dictionary, accessible by key

    Raises
    ------
    ValueError
        if the dataset name is invalid
    """

    if name == "Titanic":
        return _get_titanic_data()
    elif name == "Boston":
        return _load_boston_data()
    elif name == "cancer":
        return _get_cancer_data()
    elif name == "housing":
        return _load_housing(as_frame=False)
    else:
        raise ValueError(
            "`name should be in ['Titanic', 'Boston', 'cancer', 'housing']`"
        )


def _generated_corr_dataset_regr(size=1000):
    """Generate artificial dataset for regression tasks. Some columns are
    correlated, have no variance, large cardinality, numerical and categorical.

    Parameters
    ----------
    size : int, optional
        number of rows to generate, by default 1000

    Returns
    -------
    pd.DataFrame, pd.Series, pd.Series
        the predictors matrix, the target and the weights
    """
    # weights
    w = np.random.beta(a=1, b=0.5, size=size)
    # fixing the seed and the target
    np.random.seed(42)
    sigma = 0.2
    y = np.random.normal(1, sigma, size)
    z = y - np.random.normal(1, sigma / 5, size) + np.random.normal(1, sigma / 5, size)
    X = np.zeros((size, 13))

    # 5 relevant features, with positive and negative correlation to the target
    X[:, 0] = z
    X[:, 1] = y * np.abs(np.random.normal(0, sigma * 2, size)) + np.random.normal(
        0, sigma / 10, size
    )
    X[:, 2] = -y + np.random.normal(0, sigma, size)
    X[:, 3] = y**2 + np.random.normal(0, sigma, size)
    X[:, 4] = np.sqrt(y) + np.random.gamma(1, 0.2, size)

    # 5 irrelevant features
    X[:, 5] = np.random.normal(0, 1, size)
    X[:, 6] = np.random.poisson(1, size)
    X[:, 7] = np.random.binomial(1, 0.3, size)
    X[:, 8] = np.random.normal(0, 1, size)
    X[:, 9] = np.random.poisson(1, size)
    # zero variance
    X[:, 10] = np.ones(size)
    # high cardinality
    half_size = int(size / 2)
    X[:, 11] = np.concatenate(
        [
            np.arange(start=0, stop=half_size, step=1),
            np.arange(start=0, stop=size - half_size, step=1),
        ]
    )
    # a lot of missing values
    idx_nan = np.random.choice(size, int(round(size / 2)), replace=False)
    X[:, 12] = y**3 + np.abs(np.random.normal(0, 1, size))
    X[idx_nan, 12] = np.nan

    # make  it a pandas DF
    column_names = ["var" + str(i) for i in range(13)]
    column_names[11] = "dummy_cat"
    X = pd.DataFrame(X)
    X.columns = column_names
    X["dummy_cat"] = X["dummy_cat"].astype("category")
    # low cardinality
    nice_guys = [
        "Rick",
        "Bender",
        "Cartman",
        "Morty",
        "Fry",
        "Vador",
        "Thanos",
        "Bejita",
        "Cell",
        "Tinkywinky",
        "Lecter",
        "Alien",
        "Terminator",
        "Drago",
        "Dracula",
        "Krueger",
        "Geoffrey",
        "Goldfinder",
        "Blackbeard",
        "Excel",
        "SAS",
        "Bias",
        "Variance",
        "Scrum",
        "Human",
        "Garry",
        "Coldplay",
        "Imaginedragons",
        "Platist",
        "Creationist",
        "Gruber",
        "KeyserSoze",
        "Luthor",
        "Klaue",
        "Bane",
        "MarkZ",
    ]
    X["nice_guys"] = np.random.choice(nice_guys, X.shape[0])

    return X, y, w


def _generated_corr_dataset_classification(size=1000):
    """Generate artificial dataset for classification tasks. Some columns are
    correlated, have no variance, large cardinality, numerical and categorical.

    Parameters
    ----------
    size : int, optional
        number of rows to generate, by default 1000

    Returns
    -------
    pd.DataFrame, pd.Series, pd.Series
        the predictors matrix, the target and the weights
    """
    # weights
    w = np.random.beta(a=1, b=0.5, size=size)
    # fixing the seed and the target
    np.random.seed(42)
    y = np.random.binomial(1, 0.5, size)
    X = np.zeros((size, 13))

    z = y - np.random.binomial(1, 0.1, size) + np.random.binomial(1, 0.1, size)
    z[z == -1] = 0
    z[z == 2] = 1

    # 5 relevant features, with positive and negative correlation to the target
    X[:, 0] = z
    X[:, 1] = y * np.abs(np.random.normal(0, 1, size)) + np.random.normal(0, 0.1, size)
    X[:, 2] = -y + np.random.normal(0, 1, size)
    X[:, 3] = y**2 + np.random.normal(0, 1, size)
    X[:, 4] = np.sqrt(y) + np.random.binomial(2, 0.1, size)

    # 5 irrelevant features
    X[:, 5] = np.random.normal(0, 1, size)
    X[:, 6] = np.random.poisson(1, size)
    X[:, 7] = np.random.binomial(1, 0.3, size)
    X[:, 8] = np.random.normal(0, 1, size)
    X[:, 9] = np.random.poisson(1, size)
    # zero variance
    X[:, 10] = np.ones(size)
    # high cardinality
    X[:, 11] = np.arange(start=0, stop=size, step=1)
    # a lot of missing values
    idx_nan = np.random.choice(size, int(round(size / 2)), replace=False)
    X[:, 12] = y**3 + np.abs(np.random.normal(0, 1, size))
    X[idx_nan, 12] = np.nan

    # make  it a pandas DF
    column_names = ["var" + str(i) for i in range(13)]
    column_names[11] = "dummy"
    X = pd.DataFrame(X)
    X.columns = column_names
    X["dummy"] = X["dummy"].astype("category")

    nice_guys = [
        "Rick",
        "Bender",
        "Cartman",
        "Morty",
        "Fry",
        "Vador",
        "Thanos",
        "Bejita",
        "Cell",
        "Tinkywinky",
        "Lecter",
        "Alien",
        "Terminator",
        "Drago",
        "Dracula",
        "Krueger",
        "Geoffrey",
        "Goldfinder",
        "Blackbeard",
        "Excel",
        "SAS",
        "Bias",
        "Variance",
        "Scrum",
        "Human",
        "Garry",
        "Coldplay",
        "Imaginedragons",
        "Platist",
        "Creationist",
        "Gruber",
        "KeyserSoze",
        "Luthor",
        "Klaue",
        "Bane",
        "MarkZ",
    ]
    X["nice_guys"] = np.random.choice(nice_guys, X.shape[0])

    return X, y, w
